/* dmzSetup_ipv6.js - API library for dmzSetup IPv6 configuration */
/*
Copyright (c) 2005-2013 TeamF1, Inc.
All rights reserved.
*/
/*
modification history 
------------------------
01a,22AUG11,mvb  written. 
*/

function dmzSetupV6Validate ()
    {  
    var txtFieldIdArr = new Array ();
    txtFieldIdArr[0] = "txtLanSetupIpAddr,Please enter a valid IPv6 Address";
    txtFieldIdArr[1] = "txtPrefixLen,Please enter a valid Prefix Length";
    txtFieldIdArr[2] = "txtDhcpDomainName,Please enter a valid Domain Name";
    txtFieldIdArr[3] = "txtSvrPref,Please enter a valid Server Preference";
    txtFieldIdArr[4] = "txtPriDnsServer,Please enter a valid Primary DNS Server";
    txtFieldIdArr[5] = "txtDhcpLeaseTime,Please enter a valid DHCP lease time";

    if (txtFieldArrayCheck (txtFieldIdArr) == false)
         return false;
    
    if (ipv6Validate ('txtLanSetupIpAddr', false, true, '') == false)
	    return false;
	    
    var obj = document.getElementById('txtPrefixLen');	
    if(numericValueRangeCheck (obj, '', '', 0, 128, true, 'Invalid prefix length', '') == false)
        return false;
	
	if(!document.getElementById('txtDhcpDomainName').disabled)
        {
        if (alphaNumericValueCheck ("txtDhcpDomainName", ".-", "Invalid domain name: ") == false)
		return false;
        }
	
	var svrPrefObj = document.getElementById ('txtSvrPref');
    if (!svrPrefObj || svrPrefObj.disabled) return;
    if (numericValueRangeCheck (svrPrefObj, '', '', 0, 255, true, '', '') == false)
			return false;
    		
    if(!(document.getElementById('txtPriDnsServer').disabled))
        {
        if (ipv6Validate ('txtPriDnsServer', false, true, '') == false)
            return false;
        }
	    
	if(!(document.getElementById('txtSecDnsServer').disabled) && document.getElementById('txtSecDnsServer').value != '')
        {    
	    if (ipv6Validate ('txtSecDnsServer', false, true, '') == false)
	        return false;
        }
 
    var leaseTimeObj = document.getElementById ('txtDhcpLeaseTime');
    if (!leaseTimeObj) return;
    if (!leaseTimeObj.disabled)
        if(numericValueRangeCheck (leaseTimeObj, '', '', 1, 604800, true, 'Invalid Lease Time:', 'Seconds') == false)
            return false;
            
    return true;
    }

function dmzSetupV6Init ()
    {
    checkPort_ipv6();     
    }

function dmzPortV6EnableCheck ()
    {
    var selValue= radioCheckedValueGet ('ConfigPort.LogicalIfName');
    var hddhcpEnble = document.getElementById('hddhcpEnble').value;
    if (!selValue) return;
    if(selValue == "LAN")
    	{
        fieldStateChangeWr ('','tblDMZPortConf tblDhcpServerConf','','');
        }
    else
    	{
        fieldStateChangeWr ('','','','tblDMZPortConf tblDhcpServerConf');
        dmzDhcpPortV6Check ();
        }
    }

function addrPoolTableEnable ()
    {
    var selValue= radioCheckedValueGet ('ConfigPort.LogicalIfName');
    if (!selValue) return;
    if(selValue == "LAN")
        {
        fieldStateChangeWr ('','tblv6Pools','','');
        }
    else
        {
        fieldStateChangeWr ('','','','tblv6Pools');
        }
    }

function checkPort_ipv6()
    {
    var dmzPortObj = document.getElementById('hddmzPort');
    var ipModeObj = document.getElementById('ipModeType');
    if(dmzPortObj)
        var dmzPortObjVal = dmzPortObj.value;
    if(ipModeObj)
        var ipModeObjVal = ipModeObj.value;
    if(dmzPortObjVal == "WAN2" || ipModeObjVal != 3)
        {
        fieldStateChangeWr('','divDmzSetUp','','')
        }
    else
        {
        dmzPortV6EnableCheck ();
        }
    }
       
    
function dmzDhcpPortV6Check ()
    {
    var dhcpSelValue = comboSelectedValueGet ('selDhcpStatus');
    if (!dhcpSelValue) return;    
	switch (parseInt(dhcpSelValue, 10))
        {
        case 0:    /* Disable DHCP Server */
            fieldStateChangeWr ('txtDhcpDomainName txtDhcpLeaseTime txtSvrPref selDnsSvrs txtPriDnsServer txtSecDnsServer selDhcpMode', '', '', ''); 
            break;
        case 1:    /* Enable DHCP Server */
            fieldStateChangeWr ('', '', 'txtDhcpDomainName txtSvrPref selDnsSvrs txtDhcpLeaseTime txtPriDnsServer txtSecDnsServer selDhcpMode', '');
            checkDns();
            break;   
        }
    } 

function checkDns()
    {
    var selValue=parseInt(comboSelectedValueGet("selDnsSvrs"),10)
    if(selValue == 1 || selValue == 2)
        {
        fieldStateChangeWr("txtPriDnsServer txtSecDnsServer","","","");
        }
    else
        {
        fieldStateChangeWr("","","txtPriDnsServer txtSecDnsServer","");
        }
    }
